/*
   2111 memory chip tester - For Arduino UNO

   Adapted from original 2114 testing code
   written in 2018 by Danjovic
   and released by him under GPL V2.0
   
   Pin assignment:
   2111  AVR   Arduino Uno
    A0   PC0   A0
    A1   PC1   A1
    A2   PC2   A2
    A3   PC3   A3
    A4   PC4   A4
    A5   PC5   A5
    A6   PB0    8
    A7   PB1    9

    /CS2 PB2    10
    OD   PB3    11    
    /WE  PB4    12
    /CS  PB5    13
    IO1  PD4    4
    IO2  PD5    5
    IO3  PD6    6
    IO4  PD7    7

    Vcc         5V
    Vss         GND
*/

#include<avr/io.h>

#define CS  0b00100000
#define WE  0b00010000
#define OD  0b00001000
#define CS2 0b00000100

#define disable_2111()     PORTB |= ( CS + CS2 + WE + OD )

#define enable_CS_2111()   PORTB &= ~CS
#define disable_CS_2111()  PORTB |=  CS

#define enable_CS2_2111()  PORTB &= ~CS2
#define disable_CS2_2111() PORTB |=  CS2

#define enable_WE_2111()   PORTB &= ~WE
#define disable_WE_2111()  PORTB |=  WE

#define enable_OD_2111()   PORTB &= ~OD
#define disable_OD_2111()  PORTB |=  OD

#define float_data_lines_2111() do { DDRD &= 0x0f; PORTD |= 0xf0; } while (0)
#define activate_data_lines_2111() DDRD |= 0xf0

inline void Set_address_2111 (uint16_t address)
 {
 disable_2111() ;
 PORTC = (uint8_t)(address & (0x3f));       //  Set low address pins (a0..a5)    
 PORTB = ((uint8_t)((address >> 6) & 0x03) | 0x3C); //  Set high address pins (a6..a7) and force /CE, /CE2, OD, R/W pins high.
 }

inline void Set_data_2111 ( uint8_t data)
 {
 PORTD &= 0x0f;                             // clear upper bits
 PORTD |= (uint8_t)((data << 4) & 0xf0);    // set data in upper four bits
 }

inline uint8_t Read_data_2111(void )
 {
 return ((PIND >> 4 ) & 0x0f );
 }

inline void Write_2111 (uint16_t address, uint8_t data)
 {
 disable_2111();
 Set_address_2111 (address);

 Set_data_2111 ( data );
 enable_CS2_2111();
 enable_CS_2111();
 enable_WE_2111();

 activate_data_lines_2111();
 // Wait 4 cycles -> 250ns @ 16MHz
 asm ("NOP");
 asm ("NOP");
 asm ("NOP");
 asm ("NOP");
 disable_2111();

 float_data_lines_2111();  
 }

inline uint8_t Read_2111 (uint16_t address)
 {
 uint8_t data;
 disable_2111();
 float_data_lines_2111();
 Set_address_2111 (address);
 disable_WE_2111();
 enable_CS_2111();
 enable_CS2_2111();
 enable_OD_2111();
 // Wait 4 cycles -> 250ns @ 16MHz
 asm ("NOP");
 asm ("NOP");
 asm ("NOP");
 asm ("NOP");

 data = Read_data_2111();
 disable_OD_2111();
 disable_CS_2111();
 disable_CS2_2111();
 return data;
 }

void   Perform_test_2111( uint8_t pattern)
 {
 uint16_t address;

  // Write test pattern
  for (address = 0; address < 256; address++)
    {
    Write_2111 (address, pattern);
    }
   
  // Read back test pattern
  for (address = 0; address < 256; address++)
    {
    if (Read_2111(address) != pattern)
      {
      Serial.print ("Fail at ");
      Serial.print (address, DEC);
      break;
      }
    }

 if (address == 256) Serial.print ("OK");
 }

void setup()
  {
  // Initialize I/O
  DDRB = 0x3F;  // All outputs
  DDRC = 0x3F;  // All outputs
  DDRD = 0x00;  // All inputs
  PORTB = 0b00100000 | 0b00010000 | 0b00001000 | 0b00000100; // signals /CS and /CS2 and /WE and /OD high
  PORTC = 0;
  PORTD = 0xF0; // Pullups on D4..D7

  // Initialize serial port
  Serial.begin(9600);
  Serial.print("\nTest 2111");
 }

void loop()
 {
  for (;;)
    {
    Serial.print("\n\nStarting...");
    // Iterate through memory with different patterns
    Serial.print ("\nTest pattern 0000: ");
    Perform_test_2111(0x00);
    Serial.print ("\nTest pattern 0101: ");
    Perform_test_2111(0x5);
    Serial.print ("\nTest pattern 1010: ");
    Perform_test_2111(0xA);
    Serial.print ("\nTest pattern 1111: ");
    Perform_test_2111(0xF);
    Serial.print ("\nTest pattern 0001: ");
    Perform_test_2111(0x01);
    Serial.print ("\nTest pattern 0010: ");
    Perform_test_2111(0x2);
    Serial.print ("\nTest pattern 0100: ");
    Perform_test_2111(0x4);
    Serial.print ("\nTest pattern 1000: ");
    Perform_test_2111(0x8);
    delay(500);
    }
 }
